import numpy as np
from . import utils
from sklearn.metrics.pairwise import euclidean_distances

def calinski_harabasz_template(X, label, normalize):



	n_clusters = len(np.unique(label))
	n_samples = X.shape[0]
	n_features = X.shape[1]
	centroids = np.zeros((n_clusters, n_features))
	for i in range(n_clusters):
		centroids[i, :] = utils.centroid(X[label == i, :])
	entire_centroid = utils.centroid(X)
	compactness = 0
	separability = 0	
	for i in range(n_clusters):
		compactness += np.sum(np.square(X[label == i, :] - centroids[i, :]))
		if (normalize):
			separability += np.sum(np.square(centroids[i, :] - entire_centroid)) * (X[label == i, :].shape[0] / n_samples)
		else:
			separability += np.sum(np.square(centroids[i, :] - entire_centroid)) * (X[label == i, :].shape[0])
	result =  (separability *  (n_samples - n_clusters)) /(compactness * (n_clusters - 1)) 
	return result


def calinski_harabasz_shift_template(X, label, normalize):
	n_clusters = len(np.unique(label))
	n_samples = X.shape[0]
	n_features = X.shape[1]

	std = np.std(np.sum(np.square(X - utils.centroid(X)), axis=1))

	centroids = np.zeros((n_clusters, n_features))
	for i in range(n_clusters):
		centroids[i, :] = utils.centroid(X[label == i, :])

	entire_centroid = utils.centroid(X)

	compactness = 0
	separability = 0	
	for i in range(n_clusters):
		compactness += np.sum((np.sum(np.square(X[label == i, :] - centroids[i, :]), axis=1)))
		separability += np.sum(np.square(centroids[i, :] - entire_centroid)) * X[label == i, :].shape[0]

	compactness  /= n_samples

	separability /= n_clusters
	compactness   = np.exp(compactness / std)
	
	separability = separability / std
	invariance_term = np.exp(np.mean(np.sum(np.square(X - entire_centroid), axis=1)) / std)
	
	if (normalize):
		result = (separability * invariance_term) / (compactness * n_samples)
	else:
		result = (separability * invariance_term) / compactness 
	return result

def calinski_harabasz_shift_exp_template(X, normalize=False):
	n_samples = X.shape[0]
	std = np.std(np.sum(np.square(X - utils.centroid(X)), axis=1))
	entire_centroid = utils.centroid(X)

	compactness = np.sum((np.sum(np.square(X - entire_centroid), axis=1))) / n_samples
	compactness = np.exp(compactness / std)
	separability = 0
	invariance_term = (np.exp(np.mean(np.sum(np.square(X - entire_centroid), axis=1)) / std))


	if normalize:
		return (separability * invariance_term) / (compactness * n_samples)
	else: 
		return (separability * invariance_term) / compactness



def calinski_harabasz_dcal(X, labels):
	return calinski_harabasz_template(X, labels, normalize=True)

def calinski_harabasz_range(X, label, k=0.0015492533420772602):
	orig = calinski_harabasz(X, label)
	orig_logistic = 1 / (1 + np.exp(-k * orig))
	e_val_logistic = 0.5
	return (orig_logistic - e_val_logistic) / (1 - e_val_logistic)

def calinski_harabasz_shift(X, labels):
	return calinski_harabasz_shift_template(X, labels, normalize=False)

def calinski_harabasz_dcal_shift(X, labels):
	return calinski_harabasz_shift_template(X, labels, normalize=True)

def calinski_harabasz_dcal_range(X, labels, k=1.5498422867781867):
	orig = calinski_harabasz_dcal(X, labels)
	orig_logistic = 1 / (1 + np.exp(-k * orig))
	e_val_logistic = 0.5
	return (orig_logistic - e_val_logistic) / (1 - e_val_logistic)

def calinski_harabasz_shift_range(X, labels, k=0.004393626999113602):
	orig = calinski_harabasz_shift(X, labels)
	orig_logistic = 1 / (1 + np.exp(-k * orig))
	e_val = 0
	e_val_logistic = 1 / (1 + np.exp(-k * e_val))

	if (e_val_logistic == 1):
		return 0
	return (orig_logistic - e_val_logistic) / (1 - e_val_logistic)

def calinski_harabasz_dcal_shift_range(X, labels, k):

	orig = calinski_harabasz_dcal_shift(X, labels)
	orig_logistic = 1 / (1 + np.exp(-k * orig))
	e_val = 0
	e_val_logistic = 1 / (1 + np.exp(-k * e_val))

	if (e_val_logistic == 1):
		return 0

	return (orig_logistic - e_val_logistic) / (1 - e_val_logistic)


def calinski_harabasz_shift_range_class(X, label, k):
	return utils.pairwise_computation_k(X, label, k, calinski_harabasz_dcal_shift_range)



def calinski_harabasz_adjusted(X, labels, k):
	return calinski_harabasz_shift_range_class(X, labels, k)


def calinski_harabasz(X, labels, adjusted=False, k=4.432010535838295):
	labels = utils.change_label_to_int(labels)
	utils.sanity_check(X, labels)
	if adjusted:
		return calinski_harabasz_adjusted(X, labels, k)
	else:
		return calinski_harabasz_template(X, labels, normalize=False)


from sklearn.metrics import silhouette_score
import numpy as np
from . import utils

from scipy.spatial.distance import cdist

def silhouette_original(X, labels):
	return silhouette_score(X, labels)





def a_exp_value(X, labels, curr_label, index, std):
	curr_cluster = X[labels == curr_label, :]
	curr_point   = X[index]
	dist_mean = np.sum(np.sqrt(np.sum(np.square(curr_cluster - curr_point), axis=1))) / (curr_cluster.shape[0] - 1)

	return dist_mean

def b_exp_value(X, labels, curr_label, index, n_clusters, std):
	curr_point = X[index]
	min_dist_mean = np.inf
	cluster_size = np.inf
	for cluster_label in range(n_clusters):
		if cluster_label != curr_label:
			cluster = X[labels == cluster_label, :]
			dist_mean = np.sum(np.sqrt(np.sum(np.square(cluster - curr_point), axis=1))) / cluster.shape[0]
			if dist_mean < min_dist_mean:
				min_dist_mean = dist_mean
	
	return min_dist_mean

def silhouette_shift(X, labels):
	n_clusters = len(np.unique(labels))
	n_samples = X.shape[0]

	std = np.std(np.sqrt(np.sum(np.square(X - utils.centroid(X)), axis=1)))

	silhouette_list_zero = []
	silhouette_list_first = []
	for i in range(n_samples):
		a_val = a_exp_value(X, labels, labels[i], i, std)
		b_val = b_exp_value(X, labels, labels[i], i, n_clusters, std)
		denominator = np.exp(b_val / std) - np.exp(a_val / std)
		numerator = np.exp(max(a_val, b_val) / std)
		
		if (labels[i] == 0):
			silhouette_list_zero.append(denominator / numerator)
		else:
			silhouette_list_first.append(denominator / numerator)
	value = (np.mean(silhouette_list_zero) + np.mean(silhouette_list_first)) / 2
	return value



def silhouette_shift_class(X, labels):
	return utils.pairwise_computation(X, labels, silhouette_shift)


def silhouette_adjusted(X, labels):
	return silhouette_shift_class(X, labels)

def silhouette(X, labels, adjusted=False):
	labels = utils.change_label_to_int(labels)
	utils.sanity_check(X, labels)
	if adjusted:
		return silhouette_adjusted(X, labels)
	else:
		return silhouette_original(X, labels)


from sklearn.metrics import euclidean_distances
import numpy as np
from . import utils


def dunn_original(X, labels):
	"""
	Calculate the Dunn index of a clustering.
		:param X: The data matrix.
		:param label: The cluster labels.
		:return: The Dunn index
	"""
	n_clusters = len(np.unique(labels))

	max_intra_dist = -np.inf
	for i in range(n_clusters):
		curr_cluster = X[labels == i, :]
		curr_dists = euclidean_distances(curr_cluster)
		curr_min_dist = np.max(curr_dists)
		max_intra_dist = max(max_intra_dist, curr_min_dist)
	
	inter_dists = euclidean_distances(X[labels==0,:], X[labels==1,:])
	min_inter_dist = np.min(inter_dists)

	result = min_inter_dist / max_intra_dist
	return result




def dunn_range(X, labels, k=2980.573789948166):
	np.random.shuffle(labels)
	orig = dunn(X, labels)
	orig_logistic = 1 / (1 + np.exp(-k * orig))
	eval_logistic = 0.5
	return (orig_logistic - eval_logistic) / (1 - eval_logistic)


def dunn_shift(X, labels):
	n_clusters = len(np.unique(labels))

	std = np.std(np.sqrt(np.sum(np.square(X - utils.centroid(X)), axis=1)))

	max_intra_dist = -np.inf
	for i in range(n_clusters):
		curr_cluster = X[labels == i, :]
		curr_dists = euclidean_distances(curr_cluster)
		curr_min_dist = np.max(curr_dists)
		max_intra_dist = max(max_intra_dist, curr_min_dist)
	
	inter_dists = euclidean_distances(X[labels==0,:], X[labels==1,:])
	min_inter_dist = np.min(inter_dists)

	max_intra_dist_exp = np.exp(max_intra_dist / std)
	min_inter_dist_exp = np.exp(min_inter_dist / std)

	result = min_inter_dist_exp / max_intra_dist_exp
	return result


def dunn_shift_range(X, labels, k=18039.038395458672):
	orig = dunn_shift(X, labels)
	orig_logistic = 1 / (1 + np.exp(-k * orig))
	e_val_logistic = 0.5
	return (orig_logistic - e_val_logistic) / (1 - e_val_logistic)


def dunn_shift_range_class(X, labels):
	return utils.pairwise_computation(X, labels, dunn_shift_range)

def dunn_dcal(X, labels):
	n_clusters = len(np.unique(labels))

	max_avg_intra_dist = -np.inf
	for i in range(n_clusters):
		curr_cluster = X[labels == i, :]
		curr_dists = euclidean_distances(curr_cluster)
		curr_avg_dist = np.sum(curr_dists) / (curr_cluster.shape[0] * (curr_cluster.shape[0] - 1))
		max_avg_intra_dist = max(max_avg_intra_dist, curr_avg_dist)
	
	inter_dists = euclidean_distances(X[labels==0,:], X[labels==1,:])
	min_avg_inter_dist = np.mean(inter_dists)

	result = min_avg_inter_dist / max_avg_intra_dist
	return result


def dunn_dcal_exp(X):
	pairwise_dist = euclidean_distances(X)
	pairwise_dist_avg = np.sum(pairwise_dist) / (X.shape[0] * (X.shape[0] - 1))

	median = utils.geometric_median(X)
	centroid_dist = np.sqrt(np.sum(np.square(X - median), axis=1))
	centroid_dist_avg = np.sum(centroid_dist) / X.shape[0]

	return centroid_dist_avg / pairwise_dist_avg

def dunn_dcal_range(X, labels, k =1.269):
	orig = dunn_dcal(X, labels)
	orig_logistic = 1 / (1 + np.exp(-k * orig))
	e_val = dunn_dcal_exp(X)
	e_val_logistic = 1 / (1 + np.exp(-k * e_val))
	return (orig_logistic - e_val_logistic) / (1 - e_val_logistic)

def dunn_dcal_shift(X, labels):
	n_clusters = len(np.unique(labels))

	std = np.std(np.sqrt(np.sum(np.square(X - utils.centroid(X)), axis=1)))

	max_avg_intra_dist = -np.inf
	for i in range(n_clusters):
		curr_cluster = X[labels == i, :]
		curr_dists = euclidean_distances(curr_cluster)
		curr_avg_dist = np.sum(curr_dists) / (curr_cluster.shape[0] * (curr_cluster.shape[0] - 1))
		max_avg_intra_dist = max(max_avg_intra_dist, curr_avg_dist)
	
	inter_dists = euclidean_distances(X[labels==0,:], X[labels==1,:])
	min_avg_inter_dist = np.mean(inter_dists)

	max_avg_intra_dist = np.exp(max_avg_intra_dist / std)
	min_avg_inter_dist = np.exp(min_avg_inter_dist / std)

	result = min_avg_inter_dist / max_avg_intra_dist
	return result



def dunn_dcal_shift_exp(X):
	std = np.std(np.sqrt(np.sum(np.square(X - utils.centroid(X)), axis=1)))
	pairwise_dist = euclidean_distances(X)
	pairwise_dist_avg = np.sum(pairwise_dist) / (X.shape[0] * (X.shape[0] - 1))

	median = utils.geometric_median(X)
	centroid_dist = np.sqrt(np.sum(np.square(X - median), axis=1))
	centroid_dist_avg = np.sum(centroid_dist) / X.shape[0]

	centroid_dist_avg = np.exp(centroid_dist_avg / std)
	pairwise_dist_avg = np.exp(pairwise_dist_avg / std)

	return centroid_dist_avg / pairwise_dist_avg

def dunn_dcal_shift_range(X, labels, k ):
	orig = dunn_dcal_shift(X, labels)
	orig_logistic = 1 / (1 + np.exp(-k * orig))

	e_val = dunn_dcal_shift_exp(X)
	e_val_logistic = 1 / (1 + np.exp(-k * e_val))

	
	return (orig_logistic - e_val_logistic) / (1 - e_val_logistic)

def dunn_dcal_shift_range_class(X, labels, k):
	return utils.pairwise_computation_k(X, labels, k, dunn_dcal_shift_range)



def dunn_adjusted(X, labels, k):
	return dunn_dcal_shift_range_class(X, labels, k)


def dunn(X, labels, adjusted=False, k=0.40019810656179045):
	labels = utils.change_label_to_int(labels)
	utils.sanity_check(X, labels)
	if adjusted:
		return dunn_adjusted(X, labels, k)
	else:
		return dunn_original(X, labels)


import numpy as np
from . import utils
from .i_index import i_index_adjusted

def xie_beni_original(X, labels):
	'''
	Calculate the Xie Beni index of a clustering.
		:param X: The data matrix. 
		:param label: The cluster labels.
		:return: The Xie Beni index
	'''
	n_clusters = len(np.unique(labels))
	n_samples = X.shape[0]
	n_features = X.shape[1]
	
	"""
	compute the compacteness of each cluster as 
	the sum of the squared distances of each point to its centroid
	"""

	## compute the centroids of each cluster
	centroids = np.zeros((n_clusters, n_features))
	for i in range(n_clusters):
		centroids[i, :] = utils.centroid(X[labels == i, :])
	
	## compute the sum of squared distances of each point to its centroid
	compactness = 0
	for i in range(n_clusters):
		compactness += np.sum(np.square(X[labels == i, :] - centroids[i, :]))

	"""
	Compute the separability of each cluster as
	the sqaured minimum distance between centroids
	"""


	## compute the minimum distance between centroids
	min_dist = np.inf
	for i in range(n_clusters):
		for j in range(i + 1, n_clusters):
			dist = utils.euc_dis(centroids[i, :], centroids[j, :])
			if dist < min_dist:
				min_dist = dist
	
	min_dist_squared = min_dist * min_dist
	separability = n_samples * min_dist_squared


	return compactness / separability

def xie_beni(X, labels, adjusted=False, k=0.29289310158571386):
	labels = utils.change_label_to_int(labels)
	utils.sanity_check(X, labels)
	if adjusted:
		return i_index_adjusted(X, labels, k)
	else:
		return xie_beni_original(X, labels)




from .calinski_harabasz import calinski_harabasz
from .davies_bouldin import davies_bouldin
from .dunn import dunn
from .i_index import i_index
from .silhouette import silhouette
from .xie_beni import xie_beni

__all__ = [
    'calinski_harabasz',
    'davies_bouldin',
    'dunn',
    'i_index',
    'silhouette',
    'xie_beni'
]


import numpy as np
from scipy.spatial.distance import cdist, euclidean

def centroid(X):
	return np.mean(X, axis=0)

def euc_dis(vec_1, vec_2):
	return np.linalg.norm(vec_1 - vec_2)

def min_max_pairwise_distance(entire_dist):
	np.fill_diagonal(entire_dist, -np.inf)
	max_entire_dist = np.max(entire_dist)
	np.fill_diagonal(entire_dist, np.inf)
	min_entire_dist = np.min(entire_dist)

	return min_entire_dist, max_entire_dist

def min_max_dist(dist):
	min_dist = np.min(dist)
	max_dist = np.max(dist)
	return min_dist, max_dist

def pairwise_computation_k(X, labels, k, measure):
	class_num = len(np.unique(labels))
	result_pairwise = []
	for label_a in range(class_num):
		for label_b in range(label_a + 1, class_num):
			X_pair      = X[((labels == label_a) | (labels == label_b))]
			labels_pair = labels[((labels == label_a) | (labels == label_b))]

			unique_labels = np.unique(labels_pair)
			label_map = {old_label: new_label for new_label, old_label in enumerate(unique_labels)}
			labels_pair = np.array([label_map[old_label] for old_label in labels_pair], dtype=np.int32)

			score = measure(X_pair, labels_pair, k)
			result_pairwise.append(score)
	
	return np.mean(result_pairwise)


def pairwise_computation(X, labels, measure):
	class_num = len(np.unique(labels))
	result_pairwise = []
	for label_a in range(class_num):
		for label_b in range(label_a + 1, class_num):
			X_pair      = X[((labels == label_a) | (labels == label_b))]
			labels_pair = labels[((labels == label_a) | (labels == label_b))]

			unique_labels = np.unique(labels_pair)
			label_map = {old_label: new_label for new_label, old_label in enumerate(unique_labels)}
			labels_pair = np.array([label_map[old_label] for old_label in labels_pair], dtype=np.int32)

			score = measure(X_pair, labels_pair)
			result_pairwise.append(score)
	
	return np.mean(result_pairwise)

def geometric_median(X, eps=1e-5):
	y = np.mean(X, 0)

	while True:
		D = cdist(X, [y]) 
		nonzeros = (D != 0)[:, 0]

		Dinv = 1 / D[nonzeros]
		Dinvs = np.sum(Dinv)
		W = Dinv / Dinvs
		T = np.sum(W * X[nonzeros], 0)

		num_zeros = len(X) - np.sum(nonzeros)
		if num_zeros == 0:
			y1 = T
		elif num_zeros == len(X):
			return y
		else:
			R = (T - y) * Dinvs
			r = np.linalg.norm(R)
			rinv = 0 if r == 0 else num_zeros/r
			y1 = max(0, 1-rinv)*T + min(1, rinv)*y

		if euclidean(y, y1) < eps:
			return y1

		y = y1

def change_label_to_int(labels):
	
	unique_labels = np.unique(labels)
	label_map = {old_label: new_label for new_label, old_label in enumerate(unique_labels)}
	labels_int = np.array([label_map[old_label] for old_label in labels], dtype=np.int32)
	return labels_int

def sanity_check(X, labels):
	"""
	Comprehensive sanity check for clustering data and labels
	
	Parameters:
	-----------
	X : array-like, shape (n_samples, n_features)
		Data points
	labels : array-like, shape (n_samples,)
		Cluster labels for each data point
		
	Returns:
	--------
	bool : True if all checks pass
	
	Raises:
	-------
	ValueError : If any sanity check fails
	"""
	X = np.asarray(X)
	labels = np.asarray(labels)
	
	# Check if arrays are empty
	if X.size == 0:
		raise ValueError("Data array X is empty")
	if labels.size == 0:
		raise ValueError("Labels array is empty")
	
	# Check dimensions
	if X.ndim != 2:
		raise ValueError(f"X should be 2D array, got {X.ndim}D array")
	if labels.ndim != 1:
		raise ValueError(f"Labels should be 1D array, got {labels.ndim}D array")
	
	# Check if lengths match
	if X.shape[0] != len(labels):
		raise ValueError(f"Number of samples in X ({X.shape[0]}) does not match number of labels ({len(labels)})")
	
	# Check for minimum samples
	if X.shape[0] < 3:
		raise ValueError(f"Need at least 3 samples for clustering metrics, got {X.shape[0]}")
	
	# Check for valid data (no NaN or infinite values)
	if np.any(np.isnan(X)):
		raise ValueError("Data X contains NaN values")
	if np.any(np.isinf(X)):
		raise ValueError("Data X contains infinite values")
	
	# Get unique labels and their counts
	unique_labels, counts = np.unique(labels, return_counts=True)
	
	# Check if there are at least 2 clusters
	if len(unique_labels) < 2:
		raise ValueError("Number of unique labels is 1. Need at least 2 clusters for clustering metrics.")
	
	# Check for single-point clusters
	single_point_clusters = unique_labels[counts == 1]
	if len(single_point_clusters) > 0:
		raise ValueError(f"Found clusters with only 1 data point: {single_point_clusters}. "
						f"All clusters must have at least 2 data points for meaningful clustering metrics.")
	
	# Check for too many clusters (more clusters than samples is suspicious)
	if len(unique_labels) >= X.shape[0]:
		raise ValueError(f"Number of clusters ({len(unique_labels)}) should be less than number of samples ({X.shape[0]})")
	
	# Check for NaN values in labels
	if np.any(np.isnan(labels)):
		raise ValueError("Labels contain NaN values")
	
	# Check label data type (should be convertible to integers)
	try:
		labels_int = labels.astype(int)
		if not np.array_equal(labels.astype(float), labels_int.astype(float)):
			raise ValueError("Labels should be integers or convertible to integers")
	except (ValueError, OverflowError):
		raise ValueError("Labels should be numeric and convertible to integers")
	
	return True



from . import utils
import numpy as np
from scipy.spatial.distance import cdist, euclidean
from sklearn.metrics import euclidean_distances
import copy


def i_index_original(X, labels):
	n_clusters = len(np.unique(labels))
	n_samples = X.shape[0]
	n_features = X.shape[1]

	## compute the centroids of each cluster
	centroids= np.zeros((n_clusters, n_features))
	for i in range(n_clusters):
		centroids[i, :] = utils.centroid(X[labels == i, :])

	## compute the sum of distances of each point to its centroid
	dist_sum_to_centroids = 0
	for i in range(n_clusters):
		dists_squared = np.square(X[labels == i, :] - centroids[i, :])
		dists_row_sum = np.sqrt(np.sum(dists_squared, axis=1))
		dist_sum_to_centroids += np.sum(dists_row_sum)

	## compute the centroid of the whole data set
	centroid_whole = utils.centroid(X)

	## compute the sum of distances to the centroid of the whole data set
	dist_squared_whole = np.square(X - centroid_whole)
	dist_row_sum_whole = np.sqrt(np.sum(dist_squared_whole, axis=1))
	dist_sum_whole = np.sum(dist_row_sum_whole)

	### compute compactness
	compactness =  (dist_sum_whole / dist_sum_to_centroids) / n_clusters

	max_dist = 0
	for i in range(n_clusters):
		for j in range(i + 1, n_clusters):
			dist = utils.euc_dis(centroids[i, :], centroids[j, :])
			if dist > max_dist:
				max_dist = dist
	separability = max_dist

	power = 2

	return (compactness * separability) ** power



def i_index_exp(X, labels):
	n_clusters = len(np.unique(labels))
	n_samples = X.shape[0]
	n_features = X.shape[1]

	centroids= np.zeros((n_clusters, n_features))
	for i in range(n_clusters):
		centroids[i, :] = utils.centroid(X[labels == i, :])
	
	max_dist = 0
	for i in range(n_clusters):
		for j in range(i + 1, n_clusters):
			dist = utils.euc_dis(centroids[i, :], centroids[j, :])
			if dist > max_dist:
				max_dist = dist
	separability = max_dist

	return separability


def i_index_range(X, labels, k=0.1078):
	orig = i_index(X, labels)
	orig_logistic = 1 / (1 + np.exp(-k * orig))
	e_val_sum = 0
	labels_copy = copy.deepcopy(labels)
	for i in range(20):
		np.random.shuffle(labels_copy)
		e_val_sum += i_index(X, labels_copy)
	e_val = e_val_sum / 20
	e_val_logistic = 1 / (1 + np.exp(-k * e_val))
	if e_val_logistic == 1:
		return 0
	return (orig_logistic - e_val_logistic) / (1 - e_val_logistic)

def i_index_shift(X, labels):
	n_clusters = len(np.unique(labels))
	n_samples = X.shape[0]
	n_features = X.shape[1]

	std = np.std(np.sum(np.square(X - utils.centroid(X)), axis=1))

	std_sqrt = np.std(np.sqrt(np.sum(np.square(X - utils.centroid(X)), axis=1)))


	## compute the centroids of each cluster
	centroids= np.zeros((n_clusters, n_features))
	for i in range(n_clusters):
		centroids[i, :] = utils.centroid(X[labels == i, :])

	## compute the sum of distances of each point to its centroid
	dist_sum_to_centroids = 0
	for i in range(n_clusters):
		dists_squared = np.square(X[labels == i, :] - centroids[i, :])
		dist_sum_to_centroids += np.sum(dists_squared)

	avg_sum_to_centroids = dist_sum_to_centroids / n_samples
	
	## compute the centroid of the whole data set
	centroid_whole = utils.centroid(X)

	## compute the sum of distances to the centroid of the whole data set
	dist_squared_whole = np.square(X - centroid_whole)
	dist_sum_whole = np.sum(dist_squared_whole)

	avg_sum_whole = dist_sum_whole / n_samples


	## compute max distance between centroids
	max_centroid_dist = 0
	for i in range(n_clusters):
		for j in range(i + 1, n_clusters):
			dist = np.sqrt(np.sum(np.square(centroids[i, :] - centroids[j, :])))
			if dist > max_centroid_dist:
				max_centroid_dist = dist

	# exp_max_centroid_dist = np.exp(max_centroid_dist / std)
	exp_max_centroid_dist = max_centroid_dist / std_sqrt
	exp_sum_to_centroids  = np.exp(avg_sum_to_centroids / std)
	exp_sum_whole = np.exp(avg_sum_whole / std)

	results = (exp_sum_whole / exp_sum_to_centroids) * exp_max_centroid_dist

	return results 



def i_index_shift_range(X, labels, k):
	orig = i_index_shift(X, labels)
	orig_logistic = 1 / (1 + np.exp(-k * orig))
	e_val = 0
	e_val_logistic = 1 / (1 + np.exp(-k * e_val))

	return (orig_logistic - e_val_logistic) / (1 - e_val_logistic)


def i_index_shift_range_class(X, labels, k):
	return utils.pairwise_computation_k(X, labels, k, i_index_shift_range)




## same with xie_beni
def i_index_adjusted(X, labels, k):
	return i_index_shift_range_class(X, labels, k)

def i_index(X, labels, adjusted=False, k=0.29289310158571386):
	labels = utils.change_label_to_int(labels)
	utils.sanity_check(X, labels)
	if adjusted:
		return i_index_adjusted(X, labels, k)
	else:
		return i_index_original(X, labels)


from sklearn.metrics import davies_bouldin_score
import numpy as np
from . import utils

def davies_bouldin_sklearn(X, label):
	return davies_bouldin_score(X, label)

def davies_bouldin_original(X, label):

	n_clusters = len(np.unique(label))
	n_samples = X.shape[0]
	n_features = X.shape[1]

	centroids = np.zeros((n_clusters, n_features))
	for i in range(n_clusters):
		centroids[i, :] = utils.centroid(X[label == i, :])
	
	sum_score = 0
	for i in range(n_clusters):
		max_score = -1
		for j in range(n_clusters):
			if i == j:
				continue
			compactness  = np.sum(np.sqrt(np.sum(np.square(X[label == i, :] - centroids[i, :]), axis=1))) / X[label == i, :].shape[0]
			compactness += np.sum(np.sqrt(np.sum(np.square(X[label == j, :] - centroids[j, :]), axis=1))) / X[label == j, :].shape[0]
			separability = np.sqrt(np.sum(np.square(centroids[i, :] - centroids[j, :])))
			curr_score = compactness / separability
			if curr_score > max_score:
				max_score = curr_score
		sum_score += max_score
	
	## note that davies bouldin praises clustering with lower score
	## should be inversed to be consistent with other metrics
	return (sum_score / n_clusters) ** (-1)


	

def davies_bouldin_exp(X, label):
	return 0

def davies_bouldin_range(X, labels, k= 1.1583908441240243):
	orig = davies_bouldin(X, labels) 
	orig_logistic = 1 / (1 + np.exp(-k * orig))

	e_val = davies_bouldin_exp(X, labels)
	e_val_logistic = 1 / (1 + np.exp(-k * e_val))
	return (orig_logistic - e_val_logistic) / (1 - e_val_logistic)

def davies_bouldin_shift(X, labels):
	n_clusters = len(np.unique(labels))
	n_samples = X.shape[0]
	n_features = X.shape[1]

	std = np.std(np.sum(np.square(X - utils.centroid(X)), axis=1))

	centroids = np.zeros((n_clusters, n_features))
	for i in range(n_clusters):
		centroids[i, :] = utils.centroid(X[labels == i, :])
	
	entire_centroid = utils.centroid(X)
	
	sum_score = 0
	for i in range(n_clusters):
		max_score = -1
		for j in range(n_clusters):
			if i == j:
				continue
			compactness = np.exp(np.sum(np.square(X[labels == i, :] - centroids[i, :])) / (X[labels == i, :].shape[0] * std))
			compactness += np.exp(np.sum(np.square(X[labels == j, :] - centroids[j, :])) / (X[labels == j, :].shape[0] * std))
			separability = (np.sum(np.square(centroids[i, :] - centroids[j, :]))) / std
			invariance_term = np.sum(np.square(X - entire_centroid)) / (n_samples * std)
			curr_score = compactness / (separability * invariance_term)
			if curr_score > max_score:
				max_score = curr_score
		sum_score += max_score
	
	return (sum_score / n_clusters)  ** (-1)

def davies_bouldin_shift_exp(X, labels):
	return 0


def davies_bouldin_shift_range(X, labels, k=2.4831185988914117):
	orig = davies_bouldin_shift(X, labels)
	orig_logistic = 1 / (1 + np.exp(-k * orig))
	e_val = davies_bouldin_shift_exp(X, labels)
	e_val_logistic = 1 / (1 + np.exp(-k * e_val))
	return (orig_logistic - e_val_logistic) / (1 - e_val_logistic)

def davies_bouldin_shift_range_class(X, labels, k):
	return utils.pairwise_computation_k(X, labels,  k, davies_bouldin_shift_range)


def davies_bouldin_adjusted(X, labels, k):
	return davies_bouldin_shift_range_class(X, labels, k)

def davies_bouldin(X, labels, adjusted=False, k=2.4831185988914117):
	labels = utils.change_label_to_int(labels)
	utils.sanity_check(X, labels)
	if adjusted:
		return davies_bouldin_adjusted(X, labels, k)
	else:
		return davies_bouldin_original(X, labels)


from .ivm import (
    calinski_harabasz,
    davies_bouldin,
    dunn,
    i_index,
    silhouette,
    xie_beni
)

__version__ = "0.1.0"

__all__ = [
    'calinski_harabasz',
    'davies_bouldin', 
    'dunn',
    'i_index',
    'silhouette',
    'xie_beni'
]